// New projections for d3-geo-projection <https://github.com/d3/d3-geo-projection>
// by Peter Denner and Tobias Jung, 2023.
// For more information, see <https://blog.map-projections.net/the-winkel-denner-projections>
// or <https://blog.kartenprojektionen.de/die-winkel-denner-projektionen> (deutsch)
//
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// IMPORTANT: This is NOT a standalone file, it will need
// further scripts from d3-geo-projection in order to run.
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


// equirectangular projection taking a standard parallel
function equirectangularStdParRaw(phi0) {
    var cosPhi0 = cos(phi0);

    function forward(lambda, phi) {
        return [lambda * cosPhi0, phi];
    }

    forward.invert = function(x, y) {
        return [x / cosPhi0, y];
    };

    return forward;
}

function equirectangularStdPar() {
    return parallel1(equirectangularStdParRaw)
        .parallel(0)
        .scale(195.044);
}

// a projection similar to Winkel Tripel, but with
// a lesser amount of areal inflation.
// by tobias jung, who recommends to use the winkelish2
// or one of the winkel-denner prokections instead.
function winkelish1Raw(phi0) {
    var proj1 = hammerRaw(2),
        proj2 = equirectangularStdParRaw(phi0 * radians);

    var scaleFactor = halfPi / sqrt2;

    function forward(lambda, phi) {
        var xy1 = proj1(lambda, phi),
            xy2 = proj2(lambda, phi),

            x = (((scaleFactor * xy1[0]) + xy2[0]) / 2),
            y = (((scaleFactor * xy1[1]) + xy2[1]) / 2);

        return [x, y];
    }

    forward.invert = solve2d(forward);
    return forward;

}

function winkelish1(phi0) {
    if (phi0 === undefined) phi0 = acos(2 / pi) * degrees;
    return d3Geo.geoProjection(winkelish1Raw(phi0))
        .scale(158.837);
}



// A projection similar to Winkel Tripel, but with
// a lesser amount of areal inflation,
// by Tobias Jung,
// usage of the rescaled Wagner 0 suggested and implemented by Peter Denner.

function winkelish2Raw(inflation, phi0) {
    var proj1 = d3Geo.geoAitoffRaw,
        proj2 = wagner0RescaledRaw(inflation, phi0);

    function forward(lambda, phi) {
        var xy1 = proj1(lambda, phi),
            xy2 = proj2(lambda, phi),
            x = (xy1[0] + xy2[0]) / 2,
            y = (xy1[1] + xy2[1]) / 2;

        return [x, y];
    }

    forward.invert = solve2d(forward);
    return forward;
}

function winkelish2(inflation, phi0) {
    if (inflation === undefined) inflation = 50; // 73 approximates Winkel-Denner I; 100 reproduces Winkel Tripel
    if (phi0 === undefined) phi0 = acos(1 / halfPi) * degrees;
    return d3Geo.geoProjection(winkelish2Raw(inflation, phi0 * radians))
        .scale(158.837);
}

// Wagner 0 with same aspect ratio and area as equirectangular projection with standard parallel phi0
// A compromise cylindric projection,
// using Karlheinz Wagner's method (das Umbeziffern).
// Limiting variants are the cylindrical equal-area (no inflation) and
// the equirectangular (maximum inflation) projections.
// Originally by Tobias Jung, rescaled version suggested and implemented by Peter Denner.

function wagner0RescaledRaw(inflation, phi0) {
    var vratio = 2 * cos(phi0);
    // 0 <= inflation <= 99.999999
    inflation = max(inflation, 0);
    inflation = min(inflation, 100 - epsilon);
    // vratio > 0.
    // sensible values, i.e. something that renders a map which still can be
    // recognized as world map, are e.g. 0.2 <= vratio <= 10.
    vratio = max(vratio, epsilon);

    // convert values from boehm notation
    // areal inflation e.g. from 0 to 1 or 20 to 1.2:
    var vinflation = inflation / 100 + 1;
    // 60 is always used as reference parallel
    var phi1 = pi / 3,
        m2 = acos(vinflation * cos(phi1)) / phi1,
        k = sqrt(vratio * sin(m2 * halfPi) / pi),
        rescale = sqrt(m2) / k * cos(phi0),
        cx = rescale * k / sqrt(m2),
        cy = rescale * 1 / (k * sqrt(m2));


    function forward(lambda, phi) {
        var x = cx * lambda,
            y = cy * sin(m2 * phi);
        return [x, y];
    }

    forward.invert = function(x, y) {
        var lambda = x / cx,
            phi = 1 / m2 * asin(y / cy);
        return [lambda, phi];
    };

    return forward;
}

function wagner0Rescaled() {
    // default values generate ... a nice-looking variant (in tobias' opinion)
    var inflation = 50,
        ratio = 183,
        mutate = d3Geo.geoProjectionMutator(wagner0RescaledRaw),
        projection = mutate(inflation, ratio);

    projection.inflation = function(_) {
        return arguments.length ? mutate(inflation = +_, ratio) : inflation;
    };
    projection.ratio = function(_) {
        return arguments.length ? mutate(inflation, ratio = +_) : ratio;
    };

    return projection
        .scale(158.837);
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// Winkel Tripel alternatives by Peter Denner
function winkelRaw(variant, phi0) {
    var proj1 = d3Geo.geoAzimuthalEquidistantRaw;

    if (variant % 3 == 0) {
        var proj2 = d3Geo.geoAzimuthalEquidistantRaw,
            scaleFactor = 1;
    } else if (variant % 3 == 1) {
        var proj2 = d3Geo.geoAzimuthalEqualAreaRaw,
            scaleFactor = halfPi / sqrt2;
    } else if (variant % 3 == 2) {
        var proj2 = newAzimuthalRaw,
            scaleFactor = quarterPi / log(1 + sqrt2);
    }

    function forward(lambda, phi) {
        var xy1 = proj1(lambda / 2, phi),
            xy2 = proj2(lambda / 2, phi),
            x = (2 * xy1[0] + lambda * cos(phi0)) / 2;

        if (variant < 3) {
            var y = (scaleFactor * xy2[1] + phi) / 2;
        } else if (variant > 2) {
            var y = (scaleFactor * xy2[1] + xy1[1]) / 2;
        }

        return [x, y];
    }

    forward.invert = solve2d(forward);
    return forward;
}

function winkelTripel(phi0) {
    if (phi0 === undefined) phi0 = acos(1 / halfPi) * degrees;
    return d3Geo.geoProjection(winkelRaw(0, phi0 * radians))
        .scale(158.837);
}

function winkelDenner1(phi0) {
    if (phi0 === undefined) phi0 = acos(1 / halfPi) * degrees;
    return d3Geo.geoProjection(winkelRaw(1, phi0 * radians))
        .scale(158.837);
}

function winkelDenner2(phi0) {
    if (phi0 === undefined) phi0 = acos(1 / halfPi) * degrees;
    return d3Geo.geoProjection(winkelRaw(2, phi0 * radians))
        .scale(158.837);
}

function winkelDenner3(phi0) {
    if (phi0 === undefined) phi0 = acos(1 / halfPi) * degrees;
    return d3Geo.geoProjection(winkelRaw(3, phi0 * radians))
        .scale(158.837);
}

function winkelDenner4(phi0) {
    if (phi0 === undefined) phi0 = acos(1 / halfPi) * degrees;
    return d3Geo.geoProjection(winkelRaw(4, phi0 * radians))
        .scale(158.837);
}

function winkelDenner5(phi0) {
    if (phi0 === undefined) phi0 = acos(1 / halfPi) * degrees;
    return d3Geo.geoProjection(winkelRaw(5, phi0 * radians))
        .scale(158.837);
}

function newAzimuthalRaw(lambda, phi) {
    var delta = acos(cos(phi) * cos(lambda)),
        rho = 2 * log(tan((pi + delta) / 4));

    if (delta < epsilon) {
        var x = cos(phi) * sin(lambda),
            y = sin(phi);
    } else {
        var x = rho * (cos(phi) * sin(lambda)) / sin(delta),
            y = rho * sin(phi) / sin(delta);
    }

    return [x, y];
}

function newAzimuthal() {
    return d3Geo.geoProjection(newAzimuthalRaw)
        .scale(141.540);
}


// the following lines are also needed...
exports.geoEquirectangularStdPar = equirectangularStdPar;
exports.geoEquirectangularStdParRaw = equirectangularStdParRaw;
exports.geoWinkelish1Raw = winkelish1Raw;
exports.geoWinkelish1 = winkelish1;
exports.geoWinkelish2Raw = winkelish2Raw;
exports.geoWinkelish2 = winkelish2;
exports.geoWagner0RescaledRaw = wagner0RescaledRaw;
exports.geoWagner0Rescaled = wagner0Rescaled;
exports.geoWinkelDenner1 = winkelDenner1;
exports.geoWinkelDenner2 = winkelDenner2;
exports.geoWinkelDenner3 = winkelDenner3;
exports.geoWinkelDenner4 = winkelDenner4;
exports.geoWinkelDenner5 = winkelDenner5;
exports.geoWinkelTripel = winkelTripel;
exports.geoNewAzimuthal = newAzimuthal;
exports.geoNewAzimuthalRaw = newAzimuthalRaw;